import warnings

import torch

from graph_pes.atomic_graph import AtomicGraph
from graph_pes.graph_pes_model import GraphPESModel

# hide the annoying FutureWarning from e3nn
warnings.filterwarnings("ignore", category=FutureWarning, module="e3nn")

# fix e3nns torch.load without weights_only
if hasattr(torch.serialization, "add_safe_globals"):
    torch.serialization.add_safe_globals([slice])

__all__ = ["AtomicGraph", "GraphPESModel"]
__version__ = "0.2.1"



from __future__ import annotations

import torch

from graph_pes.atomic_graph import AtomicGraph, PropertyKey
from graph_pes.graph_pes_model import GraphPESModel


class UnitConverter(GraphPESModel):
    r"""
    A wrapper that converts the units of the energy, forces and stress
    predictions of an underlying model.

    Parameters
    ----------
    model
        The underlying model.
    energy_to_eV
        The conversion factor for energy, such that the
        ``model.predict_energy(graph) * energy_to_eV`` gives the
        energy prediction in eV.
    length_to_A
        The conversion factor for length, such that the
        ``model.predict_forces(graph) * (energy_to_eV / length_to_A)``
        gives the force prediction in eV/Å.
    """

    def __init__(
        self, model: GraphPESModel, energy_to_eV: float, length_to_A: float
    ):
        super().__init__(
            cutoff=model.cutoff.item(),
            implemented_properties=model.implemented_properties,
        )
        self._model = model
        self._energy_to_eV = energy_to_eV
        self._length_to_A = length_to_A

    def forward(self, graph: AtomicGraph) -> dict[PropertyKey, torch.Tensor]:
        predictions = self._model(graph)
        for key in predictions:
            if key in ["energy", "virial"]:
                predictions[key] *= self._energy_to_eV
            elif key == "forces":
                predictions[key] *= self._energy_to_eV / self._length_to_A
            elif key == "stress":
                predictions[key] *= self._energy_to_eV / self._length_to_A**3

        return predictions



from __future__ import annotations

import torch

from ..atomic_graph import (
    AtomicGraph,
    PropertyKey,
    neighbour_distances,
    sum_over_central_atom_index,
    sum_over_neighbours,
)
from ..graph_pes_model import GraphPESModel
from ..utils.threebody import triplet_bond_descriptors


class StillingerWeber(GraphPESModel):
    r"""
    The `Stillinger-Weber potential <https://journals.aps.org/prb/abstract/10.1103/PhysRevB.31.5262>`__
    predicts the total energy as a sum over two-body and three-body interactions:

    .. math::

        E = \epsilon \sum_i \sum_{j>i} \varphi_2(r_{ij}) + \zeta \epsilon  \cdot \sum_i \sum_{j \ne i} \sum_{k>j} \varphi_3(r_{ij}, r_{ik}, \theta_{ijk})

    where:

    .. math::

        \varphi_2(r) = A \left[ B \left( \frac{\sigma}{r} \right)^p - \left( \frac{\sigma}{r} \right)^q \right] \exp\left( \frac{\sigma}{r-a\sigma} \right)

    .. math::

        \varphi_3(r, s, \theta) = [\cos \theta - \cos \theta_0]^2 \exp\left( \frac{\gamma \sigma}{r - a\sigma} \right) \exp\left( \frac{\gamma \sigma}{s - a\sigma} \right)

    Default parameters are given for the original Stillinger-Weber potential,
    and hence are appropriate for modelling Si.

    When a parameter expects a ``tuple[float, bool]``, the first element is the
    value of the parameter and the second is a boolean indicating whether the
    parameter should be trainable.

    Parameters
    ----------
    lambda_
        The weight of the three-body term. Usually in the range 20-25.
    epsilon
        The energy scale.
    sigma
        The energy minimum
    A
        The two-body prefactor
    B
        The two-body exponent
    a
        Sets the cutoff as :math:`a \sigma`
    p
        The repulsive two-body exponent
    q
        The attractive two-body exponent
    gamma
        The three-body exponent
    """  # noqa: E501

    def __init__(
        self,
        lambda_: float = 21.0,
        epsilon: float = 2.1682,
        sigma: float = 2.0951,
        A: tuple[float, bool] = (7.049556277, False),
        B: tuple[float, bool] = (0.6022245584, False),
        a: tuple[float, bool] = (1.8, False),
        p: tuple[float, bool] = (4, False),
        q: tuple[float, bool] = (0, False),
        gamma: tuple[float, bool] = (1.2, False),
    ):
        cutoff = a[0] * sigma
        super().__init__(
            cutoff,
            implemented_properties=["local_energies"],
            three_body_cutoff=cutoff,
        )

        self.lambda_ = torch.nn.Parameter(torch.tensor(lambda_))
        self.epsilon = torch.nn.Parameter(torch.tensor(epsilon))
        self.sigma = torch.nn.Parameter(torch.tensor(sigma))

        self.A: torch.Tensor
        self.B: torch.Tensor
        self.a: torch.Tensor
        self.p: torch.Tensor
        self.q: torch.Tensor
        self.gamma: torch.Tensor
        self.theta_0: torch.Tensor

        def add_param(name: str, value: float, trainable: bool = False) -> None:
            if trainable:
                self.register_parameter(
                    name, torch.nn.Parameter(torch.tensor(value).float())
                )
            else:
                self.register_buffer(name, torch.tensor(value))

        add_param("A", *A)
        add_param("B", *B)
        add_param("a", *a)
        add_param("p", *p)
        add_param("q", *q)
        add_param("gamma", *gamma)

        self.register_buffer(
            "theta_0", torch.deg2rad(torch.tensor(109.4712206344))
        )

    def forward(self, graph: AtomicGraph) -> dict[PropertyKey, torch.Tensor]:
        r_ij = neighbour_distances(graph)
        r_ij = r_ij[r_ij < self.cutoff]  # strictly less than cutoff
        x = self.sigma / r_ij
        phi_2 = (
            self.A
            * (self.B * x**self.p - x**self.q)
            * torch.exp(self.sigma / (r_ij - self.cutoff))
        ) / 2  # divide by 2 to account for double counting
        local_e = sum_over_neighbours(phi_2, graph)

        triplet_idxs, theta, r_ij, r_ik = triplet_bond_descriptors(graph)
        # strictly less than three body cutoff
        mask = (r_ij < self.three_body_cutoff) & (r_ik < self.three_body_cutoff)
        triplet_idxs = triplet_idxs[mask]
        theta = theta[mask]
        r_ij = r_ij[mask]
        r_ik = r_ik[mask]

        phi_3 = self.lambda_ * (torch.cos(theta) - torch.cos(self.theta_0)) ** 2
        phi_3 *= torch.exp(self.gamma * self.sigma / (r_ij - self.cutoff))
        phi_3 *= torch.exp(self.gamma * self.sigma / (r_ik - self.cutoff))
        phi_3 /= 2  # double counting

        local_e += sum_over_central_atom_index(phi_3, triplet_idxs[:, 0], graph)

        return {"local_energies": local_e * self.epsilon}

    @classmethod
    def monatomic_water(cls) -> StillingerWeber:
        r"""
        The Stillinger-Weber potential for `monatomic water <https://pubs.acs.org/doi/10.1021/jp805227c>`__
        with :math:`\epsilon = 0.268381`, :math:`\sigma = 2.3925`, and
        :math:`\lambda = 23.15`.

        This potential expects as input a structure containing just the oxygen
        atoms.
        """
        return cls(epsilon=0.268381, sigma=2.3925, lambda_=23.15)



from ._mace import egret, go_mace_23, mace_mp, mace_off
from ._mattersim import mattersim
from ._orb import orb_model

__all__ = [
    "egret",
    "go_mace_23",
    "mace_mp",
    "mace_off",
    "mattersim",
    "orb_model",
]



from __future__ import annotations

import torch
from ase.data import atomic_numbers

from ..atomic_graph import (
    AtomicGraph,
    PropertyKey,
    neighbour_distances,
    sum_over_central_atom_index,
)
from ..graph_pes_model import GraphPESModel
from ..utils.misc import uniform_repr
from ..utils.nn import MLP, AtomicOneHot, UniformModuleList
from ..utils.threebody import triplet_edges


class RadialExpansion(torch.nn.Module):
    def __init__(
        self,
        cutoff: float = 5.0,
        features: int = 8,
        max_power: float = 8,
        learnable_powers: bool = False,
    ):
        super().__init__()
        self.cutoff = cutoff
        beta = (max_power / 2) ** (1.0 / (features - 1))
        powers = torch.tensor(
            [2 * (beta**i) for i in range(features)], dtype=torch.float
        )
        if learnable_powers:
            self.exponents = torch.nn.Parameter(powers)
        else:
            self.register_buffer("exponents", powers)

    def forward(
        self,
        r: torch.Tensor,  # of shape (E,)
    ) -> torch.Tensor:  # of shape (E, F)
        # apply the linear function
        f = torch.clamp(2 * (1 - r / self.cutoff), min=0)  # (E,)
        # repeat f for each exponent
        f = f.view(-1, 1).repeat(1, self.exponents.shape[0])
        # raise to the power of the exponents
        f = f ** torch.clamp(self.exponents, min=2)

        return f

    @torch.no_grad()
    def __repr__(self):
        return uniform_repr(self.__class__.__name__, cutoff=self.cutoff)


class TwoBodyDescriptor(torch.nn.Module):
    """
    Two-body term of the EDDP potential.
    """

    def __init__(
        self,
        Z1: int,
        Z2: int,
        cutoff: float = 5.0,
        features: int = 8,
        max_power: float = 8,
        learnable_powers: bool = False,
    ):
        super().__init__()
        self.Z1 = Z1
        self.Z2 = Z2

        self.expansion = RadialExpansion(
            cutoff, features, max_power, learnable_powers
        )

    def forward(self, r: torch.Tensor, graph: AtomicGraph) -> torch.Tensor:
        # select the relevant edges
        i = graph.neighbour_list[0]
        j = graph.neighbour_list[1]
        mask = (graph.Z[i] == self.Z1) & (graph.Z[j] == self.Z2)
        r = r[mask]

        # expand each edge
        f_edge = self.expansion(r)

        # sum over central atom
        f_atom = sum_over_central_atom_index(f_edge, i[mask], graph)
        return f_atom


class ThreeBodyDescriptor(torch.nn.Module):
    """
    Three-body term of the EDDP potential.
    """

    def __init__(
        self,
        Z1: int,
        Z2: int,
        Z3: int,
        cutoff: float = 5.0,
        features: int = 8,
        max_power: float = 8,
        learnable_powers: bool = False,
    ):
        super().__init__()
        self.Z1 = Z1
        self.Z2 = Z2
        self.Z3 = Z3
        self.cutoff = cutoff

        self.central_atom_expansion = RadialExpansion(
            cutoff, features, max_power, learnable_powers
        )
        self.neighbour_atom_expansion = RadialExpansion(
            cutoff, features, max_power, learnable_powers
        )

    def forward(
        self, i, j, k, r_ij, r_ik, r_jk, graph: AtomicGraph
    ) -> torch.Tensor:
        mask = (
            (graph.Z[i] == self.Z1)
            & (graph.Z[j] == self.Z2)
            & (graph.Z[k] == self.Z3)
        )
        r_ij = r_ij[mask]
        r_ik = r_ik[mask]
        r_jk = r_jk[mask]

        e_ij = self.central_atom_expansion(r_ij)
        e_ik = self.central_atom_expansion(r_ik)
        e_jk = self.neighbour_atom_expansion(r_jk)

        E = r_ij.shape[0]
        F = e_ij.shape[1]
        prod = (e_ij * e_ik)[:, None, :] * e_jk[:, :, None]
        prod = prod.view(E, F * F)

        # sum over central atoms
        return sum_over_central_atom_index(prod, i[mask], graph)


class EDDP(GraphPESModel):
    r"""
    The Ephemeral Data Derived Potential (EDDP) architecture, from
    `Ephemeral data derived potentials for random structure 
    search. Phys. Rev. B 106, 014102 <https://doi.org/10.1103/PhysRevB.106.014102>`_.

    This model makes local energy predictions by featurising the local
    atomic environment and passing the result through an MLP:

    .. math::

        \varepsilon_i = \text{MLP}(\mathbf{F}_i)

    where :math:`\mathbf{F}_i` is a a concatenation of one-, two- and
    three-body features:

    .. math::

        \mathbf{F}_i = \mathbf{F}_i^{(1)} \oplus \mathbf{F}_i^{(2)}
        \oplus \mathbf{F}_i^{(3)}

    as per Eq. 14. The m'th component of the two-body feature is given by:

    .. math::

        \mathbf{F}_i^{(2)}[m] = \sum_{j \in \mathcal{N}(i)} f(r_{ij})^{p_m}

    where :math:`r_{ij}` is the distance between atoms :math:`i` and :math:`j`,
    :math:`\mathcal{N}(i)` is the set of neighbours of atom :math:`i`,
    :math:`p_m \geq 2` is some power, and :math:`f(r_{ij})` is defined as:

    .. math::

        f(r_{ij}) = \begin{cases}
            2 (1 - r_{ij} / r_{\text{cutoff}}) & \text{if } r_{ij} \leq 
            r_{\text{cutoff}} \\
            0 & \text{otherwise}
        \end{cases}
    
    as per Eq. 7. Finally, the :math:`(m, o)`'th component of the three-body
    feature is given by:

    .. math::

        \mathbf{F}_i^{(3)}[m, o] = \sum_{j \in \mathcal{N}(i)}
        \sum_{k > j \in \mathcal{N}(i)} f(r_{ij})^{p_m} f(r_{ik})^{p_m}
        f(r_{jk})^{p_o}

    If you use this model, please cite the original paper:

    .. code-block:: bibtex

        @article{Pickard-22-07,
            title = {
                Ephemeral Data Derived Potentials for Random Structure Search
            },
            author = {Pickard, Chris J.},
            year = {2022},
            journal = {Physical Review B},
            volume = {106},
            number = {1},
            pages = {014102},
            doi = {10.1103/PhysRevB.106.014102},
        }

    Parameters
    ----------
    elements:
        The elements the model will be applied to (e.g. ``["C", "H"]``).
    cutoff:
        The maximum distance between pairs of atoms to consider their
        interaction.
    features:
        The number of features used to describe the two body interactions.
    max_power:
        The maximum power of the interaction expansion for two body terms.
    mlp_width:
        The width of the MLP (i.e. number of neurons in each hidden layer).
    mlp_layers:
        The number of hidden layers in the MLP.
    activation:
        The activation function to use in the MLP.
    three_body_cutoff:
        The radial cutoff for three body interactions. If ``None``, the
        same as ``cutoff``.
    three_body_features:
        The number of features used to describe the three body interactions.
        If ``None``, the same as for the two body terms.
    three_body_max_power:
        The maximum power of the interaction expansion for three body terms.
        If ``None``, the same as for the two body terms.
    learnable_powers:
        If ``True``, the powers of the interaction expansion are learnable.
        If ``False``, the powers are fixed to the values given by ``max_power``.  
    """  # noqa: E501

    def __init__(
        self,
        elements: list[str],
        cutoff: float = 5.0,
        features: int = 8,
        max_power: float = 8,
        mlp_width: int = 16,
        mlp_layers: int = 1,
        activation: str = "CELU",
        three_body_cutoff: float | None = None,
        three_body_features: int | None = None,
        three_body_max_power: float | None = None,
        learnable_powers: bool = False,
    ):
        if three_body_cutoff is None:
            three_body_cutoff = cutoff
        if three_body_features is None:
            three_body_features = features
        if three_body_max_power is None:
            three_body_max_power = max_power

        super().__init__(
            cutoff=max(cutoff, three_body_cutoff),
            implemented_properties=["local_energies"],
            three_body_cutoff=three_body_cutoff,
        )

        self.elements = elements
        Zs = [atomic_numbers[Z] for Z in elements]

        # one body terms
        self.one_hot = AtomicOneHot(elements)

        # two body terms
        self.Z_pairs = [(Z1, Z2) for Z1 in Zs for Z2 in Zs]
        self.two_body_descriptors = UniformModuleList(
            [
                TwoBodyDescriptor(
                    Z1,
                    Z2,
                    cutoff=cutoff,
                    features=features,
                    max_power=max_power,
                    learnable_powers=learnable_powers,
                )
                for Z1, Z2 in self.Z_pairs
            ]
        )

        # three body terms
        self.Z_triplets = [
            (Z1, Z2, Z3)
            for Z1 in Zs
            for Z2 in Zs
            for Z3 in Zs
            if Z2 <= Z3  # skip identical triplets (A,B,C) and (A,C,B)
        ]
        self.three_body_descriptors = UniformModuleList(
            [
                ThreeBodyDescriptor(
                    Z1,
                    Z2,
                    Z3,
                    cutoff=three_body_cutoff,
                    features=three_body_features,
                    max_power=three_body_max_power,
                    learnable_powers=learnable_powers,
                )
                for Z1, Z2, Z3 in self.Z_triplets
            ]
        )

        # read out head
        input_features = (
            len(elements)
            + len(self.two_body_descriptors) * features
            + len(self.three_body_descriptors) * three_body_features**2
        )
        layers = [input_features] + [mlp_width] * mlp_layers + [1]
        self.mlp = MLP(layers, activation=activation)

        self.shifts = torch.nn.Parameter(torch.zeros(input_features))
        self.scales = torch.nn.Parameter(torch.ones(input_features))

    def featurise(self, graph: AtomicGraph) -> torch.Tensor:
        # one body terms
        central_atom_features = [self.one_hot(graph.Z)]

        # two body terms
        rs = neighbour_distances(graph)
        for descriptor in self.two_body_descriptors:
            central_atom_features.append(descriptor(rs, graph))

        # three body terms
        i, j, k, r_ij, r_ik, r_jk = triplet_edges(
            graph, self.three_body_cutoff.item()
        )

        for descriptor in self.three_body_descriptors:
            central_atom_features.append(
                descriptor(i, j, k, r_ij, r_ik, r_jk, graph)
            )

        # concatenate all features
        f = torch.cat(central_atom_features, dim=1)
        f = (f - self.shifts) / self.scales
        return f

    def forward(self, graph: AtomicGraph) -> dict[PropertyKey, torch.Tensor]:
        features = self.featurise(graph)
        return {"local_energies": self.mlp(features).view(-1)}

    def pre_fit(self, graphs: AtomicGraph) -> None:
        X = self.featurise(graphs)
        _mean = X.mean(dim=0)
        _std = X.std(dim=0)
        # replace nans
        _std[torch.isnan(_std) | (_std < 1e-6)] = 1
        self.shifts.data = _mean
        self.scales.data = _std



from __future__ import annotations

import torch

from graph_pes.atomic_graph import (
    AtomicGraph,
    get_vectors,
    neighbour_distances,
    number_of_atoms,
    number_of_edges,
)


def angle_spanned_by(v1: torch.Tensor, v2: torch.Tensor):
    """
    Calculate angles between corresponding vectors in two batches.

    Parameters
    ----------
    v1
        First batch of vectors, shape (N, 3)
    v2
        Second batch of vectors, shape (N, 3)

    Returns
    -------
    torch.Tensor
        Angles in radians, shape (N,)
    """
    dot_product = torch.sum(v1 * v2, dim=1)

    v1_mag = torch.linalg.vector_norm(v1, dim=1)
    v2_mag = torch.linalg.vector_norm(v2, dim=1)

    cos_angle = dot_product / (v1_mag * v2_mag)
    cos_angle = torch.clamp(cos_angle, min=-1.0 + 1e-7, max=1.0 - 1e-7)

    return torch.arccos(cos_angle)


def triplet_bond_descriptors(
    graph: AtomicGraph,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    r"""
    For each triplet :math:`(i, j, k)`, get the bond angle :math:`\theta_{jik}`
    (in radians) and the two bond lengths :math:`r_{ij}` and :math:`r_{ik}`.

    Returns
    -------
    triplet_idxs
        The triplet indices, :math:`(i, j, k)`, of shape ``(Y, 3)``.
    angle
        The bond angle :math:`\theta_{jik}`, shape ``(Y,)``.
    r_ij
        The bond length :math:`r_{ij}`, shape ``(Y,)``.
    r_ik
        The bond length :math:`r_{ik}`, shape ``(Y,)``.

    Examples
    --------
    >>> graph = AtomicGraph.from_ase(molecule("H2O"))
    >>> angle, r_ij, r_ik = triplet_bond_descriptors(graph)
    >>> torch.rad2deg(angle)
    tensor([103.9999, 103.9999,  38.0001,  38.0001,  38.0001,  38.0001])
    """

    ij, S_ij, ik, S_ik = triplet_edge_pairs(graph, graph.cutoff)  # (Y, 2)

    triplet_idxs = torch.cat([ij, ik[1, :].unsqueeze(0)], dim=0).transpose(
        0, 1
    )  # (Y, 3)

    if triplet_idxs.shape[0] == 0:
        return (
            triplet_idxs,
            torch.zeros(0, device=graph.R.device).float(),
            torch.zeros(0, device=graph.R.device).float(),
            torch.zeros(0, device=graph.R.device).float(),
        )

    v1 = get_vectors(graph, i=ij[0, :], j=ij[1, :], shifts=S_ij)
    v2 = get_vectors(graph, i=ik[0, :], j=ik[1, :], shifts=S_ik)

    return (
        triplet_idxs,
        angle_spanned_by(v1, v2),
        torch.linalg.vector_norm(v1, dim=-1),
        torch.linalg.vector_norm(v2, dim=-1),
    )


def triplet_edge_pairs(
    graph: AtomicGraph,
    three_body_cutoff: float,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    r"""
    Find all the pairs of edges, :math:`(i, j), (i, k)`, such that:

    * :math:`i, j, k \in \{0, 1, \dots, N-1\}` are indices of distinct
      (images of) atoms within the graph
    * :math:`j \neq k`
    * :math:`r_{ij} \leq` ``three_body_cutoff``
    * :math:`r_{ik} \leq` ``three_body_cutoff``

    Returns
    -------
    ij: torch.Tensor
    S_ij: torch.Tensor
    ik: torch.Tensor
    S_ik: torch.Tensor
    """

    if three_body_cutoff > graph.cutoff + 1e-6:
        raise ValueError(
            "Three-body cutoff is greater than the graph cutoff. "
            "This is not currently supported."
        )

    # check if already cached, creating the key in a torchscript compatible way
    # NB this gets added in the to_batch function, which is called on the worker
    #    threads. Since this function is slow, this speeds up training, but
    #    should not be used for MD/inference. Hence we don't cache any results
    #    to the graph within this function.

    # stupidly verbose to make torchscript happy
    cutoff_str = str(
        torch.round(torch.tensor(float(three_body_cutoff)), decimals=3).item()
    )
    key = "__threebody-" + cutoff_str
    ij = graph.other.get(key + "-ij")
    S_ij = graph.other.get(key + "-Sij")
    ik = graph.other.get(key + "-ik")
    S_ik = graph.other.get(key + "-Sik")
    if ij is not None:  # noqa: SIM102
        if S_ij is not None:  # noqa: SIM102
            if ik is not None:  # noqa: SIM102
                if S_ik is not None:  # noqa: SIM102
                    return (ij, S_ij, ik, S_ik)

    with torch.no_grad():
        edge_indexes = torch.arange(
            number_of_edges(graph), device=graph.R.device
        )

        three_body_mask = neighbour_distances(graph) < three_body_cutoff
        relevant_edge_indexes = edge_indexes[three_body_mask]
        relevant_central_atoms = graph.neighbour_list[0][relevant_edge_indexes]

        edge_pairs = []

        for i in range(number_of_atoms(graph)):
            mask = relevant_central_atoms == i
            masked_edge_indexes = relevant_edge_indexes[mask]

            # number of edges of distance <= three_body_cutoff
            # that have i as a central atom
            N = masked_edge_indexes.shape[0]
            _idx = torch.cartesian_prod(
                torch.arange(N),
                torch.arange(N),
            )  # (N**2, 2)
            _idx = _idx[_idx[:, 0] != _idx[:, 1]]  # (N**2 - N, 2)

            pairs_for_i = masked_edge_indexes[_idx]
            edge_pairs.append(pairs_for_i)

        edge_pairs = torch.cat(edge_pairs)
        if edge_pairs.shape[0] == 0:
            return (
                torch.zeros(2, 0, device=graph.R.device),
                torch.zeros(0, 3, device=graph.R.device),
                torch.zeros(2, 0, device=graph.R.device),
                torch.zeros(0, 3, device=graph.R.device),
            )

        a, b = edge_pairs[:, 0], edge_pairs[:, 1]

        return (
            graph.neighbour_list[:, a],
            graph.neighbour_cell_offsets[a],
            graph.neighbour_list[:, b],
            graph.neighbour_cell_offsets[b],
        )


def triplet_edges(
    graph: AtomicGraph,
    three_body_cutoff: float,
) -> tuple[
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
]:
    """
    Finds all ``Y`` triplets ``(i, j, k)`` such that:

    * ``i, j, k`` are indices of distinct (images of) atoms within the graph
    * ``r_{ij} <=`` ``three_body_cutoff``
    * ``r_{ik} <=`` ``three_body_cutoff``

    Returns
    -------
    i: torch.Tensor
        The central atom indices, shape ``(Y,)``.
    j: torch.Tensor
        The first paired atom indices, shape ``(Y,)``.
    k: torch.Tensor
        The second paired atom indices, shape ``(Y,)``.
    r_ij: torch.Tensor
        The bond length :math:`r_{ij}`, shape ``(Y,)``.
    r_ik: torch.Tensor
        The bond length :math:`r_{ik}`, shape ``(Y,)``.
    r_jk: torch.Tensor
        The bond length :math:`r_{jk}`, shape ``(Y,)``.
    """

    ij, S_ij, ik, S_ik = triplet_edge_pairs(graph, three_body_cutoff)

    v_ij = get_vectors(graph, i=ij[0, :], j=ij[1, :], shifts=S_ij)
    v_ik = get_vectors(graph, i=ik[0, :], j=ik[1, :], shifts=S_ik)
    v_jk = v_ik - v_ij

    r_ij = torch.norm(v_ij, dim=-1)
    r_ik = torch.norm(v_ik, dim=-1)
    r_jk = torch.norm(v_jk, dim=-1)

    return (
        ij[0, :],
        ij[1, :],
        ik[1, :],
        r_ij,
        r_ik,
        r_jk,
    )



from __future__ import annotations

import time
from pathlib import Path
from typing import Final

from graph_pes.utils.misc import silently_create_trainer

# dirty hack: just get lightning to work this out,
# and ensure no annoying printing happens
_trainer = silently_create_trainer(logger=False)

GLOBAL_RANK: Final[int] = _trainer.global_rank
WORLD_SIZE: Final[int] = _trainer.world_size
IS_RANK_0: Final[bool] = GLOBAL_RANK == 0


class Communication:
    def __init__(self, key: str, timeout_s: float = 20.0):
        if "\n" in key:
            raise ValueError("Key cannot contain newlines")
        self.key = key
        self.timeout_s = timeout_s
        self.path = Path(self.key)
        assert not self.path.exists()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # wait for all ranks to reach this point
        # so that the path and contents are not lost
        # before other ranks have read them
        _trainer.strategy.barrier(self.key)
        self.path.unlink(missing_ok=True)

    def send_to_other_ranks(self, value: str) -> None:
        assert IS_RANK_0

        # write the value to the path on rank 0
        self.path.write_text(value)

    def receive_from_rank_0(self) -> str:
        assert not IS_RANK_0

        # read the value from the path on non-0 ranks,
        # re-trying a few times until the file is found
        # in cases where rank 0 is slow

        for _ in range(int(self.timeout_s * 10)):
            if self.path.exists():
                return self.path.read_text()
            time.sleep(0.1)

        raise RuntimeError(
            f"Key {self.key} not found after {self.timeout_s}s "
            f"on rank {GLOBAL_RANK}"
        )



from __future__ import annotations

import re
import warnings
from typing import TypeVar

warnings.filterwarnings(
    "ignore",
    module="e3nn",
    message=".*you are using `torch.load`.*",
)

import pathlib

import torch

from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.utils.logger import logger

from .addition import AdditionModel
from .e3nn.mace import MACE, ZEmbeddingMACE
from .e3nn.nequip import NequIP, ZEmbeddingNequIP
from .eddp import EDDP
from .offsets import FixedOffset, LearnableOffset
from .painn import PaiNN
from .pairwise import (
    LennardJones,
    LennardJonesMixture,
    Morse,
    PairPotential,
    SmoothedPairPotential,
    ZBLCoreRepulsion,
)
from .schnet import SchNet
from .scripted import ScriptedModel
from .stillinger_weber import StillingerWeber
from .tensornet import TensorNet
from .unit_converter import UnitConverter

__all__ = [
    "AdditionModel",
    "EDDP",
    "FixedOffset",
    "LearnableOffset",
    "LennardJones",
    "LennardJonesMixture",
    "MACE",
    "Morse",
    "NequIP",
    "PaiNN",
    "PairPotential",
    "SchNet",
    "SmoothedPairPotential",
    "TensorNet",
    "ZBLCoreRepulsion",
    "ZEmbeddingMACE",
    "ZEmbeddingNequIP",
    "StillingerWeber",
    "UnitConverter",
]

MODEL_EXCLUSIONS = {
    "FixedOffset",
    "LearnableOffset",
    "AdditionModel",
    "PairPotential",
    "SmoothedPairPotential",
    "UnitConverter",
}

ALL_MODELS: list[type[GraphPESModel]] = [
    globals()[model] for model in __all__ if model not in MODEL_EXCLUSIONS
]


def load_model(path: str | pathlib.Path) -> GraphPESModel:
    """
    Load a model from a file.

    Parameters
    ----------
    path
        The path to the file.

    Returns
    -------
    GraphPESModel
        The model.

    Examples
    --------

    Use this function to load an existing model for further training using
    ``graph-pes-train``:

    .. code-block:: yaml

        model:
            +load_model:
                path: path/to/model.pt

    See :doc:`fine-tuning <../quickstart/fine-tuning>` for more details.
    """
    path = pathlib.Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Could not find model at {path}")

    model = torch.load(path, weights_only=False)

    if isinstance(model, torch.jit.ScriptModule):
        model = ScriptedModel(model)

    if not isinstance(model, GraphPESModel):
        raise ValueError(
            "Expected the loaded object to be a GraphPESModel "
            f"but got {type(model)}"
        )

    import graph_pes

    if model._GRAPH_PES_VERSION != graph_pes.__version__:
        warnings.warn(
            "You are attempting to load a model that was trained with "
            f"a different version of graph-pes ({model._GRAPH_PES_VERSION}) "
            f"than what you are currently using ({graph_pes.__version__}). "
            "We won't stop you from doing this, but it may cause issues.",
            stacklevel=2,
        )

    return model.to("cpu")


def load_model_component(
    path: str | pathlib.Path,
    key: str,
) -> GraphPESModel:
    """
    Load a component from an :class:`~graph_pes.models.AdditionModel`.

    Parameters
    ----------
    path
        The path to the file.
    key
        The key to load.

    Returns
    -------
    GraphPESModel
        The component.
    """

    base_model = load_model(path)
    if not isinstance(base_model, AdditionModel):
        raise ValueError(
            f"Expected to load an AdditionModel, got {type(base_model)}"
        )

    return base_model[key]


T = TypeVar("T", bound=torch.nn.Module)


def freeze(model: T) -> T:
    """
    Freeze all parameters in a module.

    Parameters
    ----------
    model
        The model to freeze.

    Returns
    -------
    T
        The model.
    """
    for param in model.parameters():
        param.requires_grad = False
    return model


def freeze_matching(model: T, pattern: str) -> T:
    r"""
    Freeze all parameters that match the given pattern.

    Parameters
    ----------
    model
        The model to freeze.
    pattern
        The regular expression to match the names of the parameters to freeze.

    Returns
    -------
    T
        The model.

    Examples
    --------

    Freeze all the parameters in the first layer of a MACE-MP0 model from
    :func:`~graph_pes.interfaces.mace_mp` (which have names of the form
    ``"model.interactions.0.<name>"``):

    .. code-block:: yaml

        model:
            +freeze_any_matching:
                model: +mace_mp()
                pattern: model\.interactions\.0\..*
    """
    for name, param in model.named_parameters():
        if re.match(pattern, name):
            logger.info(f"Freezing {name}")
            param.requires_grad = False
    return model


def freeze_any_matching(model: T, patterns: list[str]) -> T:
    r"""
    Freeze all parameters that match any of the given patterns.

    Parameters
    ----------
    model
        The model to freeze.
    patterns
        The patterns to match.

    Returns
    -------
    T
        The model.
    """
    for pattern in patterns:
        freeze_matching(model, pattern)
    return model


def freeze_all_except(model: T, pattern: str | list[str]) -> T:
    r"""
    Freeze all parameters in a model except those matching a given pattern.

    Parameters
    ----------
    model
        The model to freeze.
    pattern
        The pattern/s to match.

    Returns
    -------
    T
        The model.

    Examples
    --------

    Freeze all parameters in a MACE-MP0 model from
    :func:`~graph_pes.interfaces.mace_mp` except those in the
    read-out heads:

    .. code-block:: yaml

        model:
            +freeze_all_except:
                model: +mace_mp()
                pattern: model\.readouts.*
    """
    freeze(model)

    if isinstance(pattern, str):
        pattern = [pattern]
    for name, param in model.named_parameters():
        if any(re.match(p, name) for p in pattern):
            param.requires_grad = True

    return model
